!$
!$=head1 NAME
!$
!$Stack3d - Stack a SEPlib dataset
!$
!$Infill  - Infill  a SEP3D dataset
!$
!$Stack - Stack a SEP regular dataset
!$
!$=head1 SYNOPSIS
!$
!$Stack3d  <in.H >out.H
!$
!$=head1 INPUT PARAMETERS
!$
!$=over 4
!$
!$=item axes- integer
!$
!$      [2]  Compress axes 2 through ....
!$
!$       Defaults to 2 for Stack and Stack3d, no compression for infill
!$
!$
!$=item maxsize- integer
!$
!$      [10] Maximum memory in megabytes
!$
!$=item verb- logical
!$
!$      [n]  Whether (y) or not (n) to be verbose
!$
!$=item normalize- logical
!$
!$      [y]  Whether (y) or not (n) to normalize by the
!$      number of traces in bin
!$
!$=back
!$
!$=head1 DESCRIPTION
!$
!$ Stack3d a sep3d dataset producing a normalized sep77 cube compressing grid axes
!$
!$ Infill3d a sep3d dataset producing a normalized sep77 cube
!$
!$ Stack - Compress axis 2 of regular. This is only for backward compatibility
!$
!$     Stack's options are a subset of Stack3d option
!$
!$=head1 NOTES
!$
!$ axes is the last axis that you want to stack over.
!$
!$ axes=2 mean stack over the 2nd axis, axes=3 mean the second and third axis
!$
!$
!$
!$=head1 CATEGORY
!$
!$B<util/cube>
!$
!$=cut
!$
!
!
!>
!-------------------------------------------------
!
!Author: Robert Clapp, ESMB 463, 7230253
!
!Date Created:Tue Apr  7  9:11:52 PDT 1998
!Modified    Bob  Jul 6 Rewrote in sep3d
!
!Purpose:
!
program Stack3d 
  use sep3d_struct_mod
  use helixcartmod
  use sep
!DATA VARIABLES
  implicit none
  type(sep3d)         ::  in,out
  real,allocatable    ::  input(:,:),output(:,:)
  complex,allocatable    ::  inputc(:,:),outputc(:,:)
  integer,allocatable ::  n(:)
  real   ,allocatable ::  o(:),d(:)
  character(len=128),allocatable :: label(:),unit(:)
  integer             ::  ndim,esize
!PARAMETERS
  integer             ::  axes,maxsize
  logical             ::  verbose,normalize
!EXTERNALS
  integer             ::  sreed,srite
!INTERNAL VARIABLES
  integer             ::  ntr,read,block,nh
  integer             ::  tempi,temp2i
  integer, allocatable :: norm(:)
  integer             ::  i3,i2,ierr,ipct_print,ipct,tot
  character(len=128)  ::  progname
  integer,parameter :: big_int=selected_int_kind(11)
  integer(kind=big_int) ::  nhelix,ncompress,nbig
  integer               :: nhelix_s,ncompress_s,nbig_s
  integer,allocatable   :: coords(:,:)
  integer       :: ndone_big
  integer       :: ndo_big,nblock_big
!INITIALIZE
  call initpar()
call sep_add_doc_line("NAME")
call sep_add_doc_line("    Stack3d - Stack a SEPlib dataset")
call sep_add_doc_line("")
call sep_add_doc_line("    Infill - Infill a SEP3D dataset")
call sep_add_doc_line("")
call sep_add_doc_line("    Stack - Stack a SEP regular dataset")
call sep_add_doc_line("")
call sep_add_doc_line("SYNOPSIS")
call sep_add_doc_line("    Stack3d <in.H >out.H")
call sep_add_doc_line("")
call sep_add_doc_line("INPUT PARAMETERS")
call sep_add_doc_line("    axes- integer")
call sep_add_doc_line("              [2]  Compress axes 2 through ....")
call sep_add_doc_line("")
call sep_add_doc_line("               Defaults to 2 for Stack and Stack3d, no compression for infill")
call sep_add_doc_line("")
call sep_add_doc_line("    maxsize- integer")
call sep_add_doc_line("              [10] Maximum memory in megabytes")
call sep_add_doc_line("")
call sep_add_doc_line("    verb- logical")
call sep_add_doc_line("              [n]  Whether (y) or not (n) to be verbose")
call sep_add_doc_line("")
call sep_add_doc_line("    normalize- logical")
call sep_add_doc_line("              [y]  Whether (y) or not (n) to normalize by the")
call sep_add_doc_line("              number of traces in bin")
call sep_add_doc_line("")
call sep_add_doc_line("DESCRIPTION")
call sep_add_doc_line("     Stack3d a sep3d dataset producing a normalized sep77 cube compressing grid axes")
call sep_add_doc_line("")
call sep_add_doc_line("     Infill3d a sep3d dataset producing a normalized sep77 cube")
call sep_add_doc_line("")
call sep_add_doc_line("     Stack - Compress axis 2 of regular. This is only for backward compatibility")
call sep_add_doc_line("")
call sep_add_doc_line("         Stack's options are a subset of Stack3d option")
call sep_add_doc_line("")
call sep_add_doc_line("NOTES")
call sep_add_doc_line("     axes is the last axis that you want to stack over.")
call sep_add_doc_line("")
call sep_add_doc_line("     axes=2 mean stack over the 2nd axis, axes=3 mean the second and third axis")
call sep_add_doc_line("")
call sep_add_doc_line("CATEGORY")
call sep_add_doc_line("    util/cube")
call sep_add_doc_line("")

  call doc('')
  call init_3d()
  call init_sepf90()
!##########################
!  READ IN THE PARAMETERS #
!##########################
  call from_param("maxsize",maxsize,10)
  call from_param("normalize",normalize,.false.)
  call from_param("verb",verbose,.false.)
  call sep_prog(progname)
  if (progname(1:8).eq."Infill3d") then
    axes=1
  else
    axes=2
  end if 
  call from_param("axes",axes,axes)
!##########################
!LETS FIGURE OUT THE INPUT#
!##########################
  call init_sep3d("in",in,"INPUT")  !Initialize SEP3D structure
  if (in%data_format(1:5).eq."FLOAT") then
    esize=4
  else if (in%data_format(1:5).eq."COMPL") then
    esize=8
  else
    call erexit("only know how to handle float data")
  end if 
!###################################
!CHECK TO SEE IF PARAMETERS LOGICAL#
!###################################
!if(axes==1) call erexit("Call Infill3d instead ")
  if (axes<1 .or. axes > in%ndims) then
    call erexit("axes out of range")
  end if
  if (verbose) then
    write(0,*) "Compressing through axes:",axes
  end if
  if ((maxsize*1000000)/esize/in%n(1)<1) then
    call erexit("maxsize to small")
  end if
!####################
!CREATE OUTPUT SPACE#
!####################
  ndim=in%ndims-axes+1
  allocate(n(ndim),o(ndim),d(ndim),label(ndim),unit(ndim))
  n(1)=in%n(1)
  o(1)=in%o(1)
  d(1)=in%d(1)
  label(1)=in%label(1)
  unit(1)=in%unit(1)
  if (ndim>1) then
    n(2:)=in%n(axes+1:)
    o(2:)=in%o(axes+1:)
    d(2:)=in%d(axes+1:)
    label(2:)=in%label(axes+1:)
    unit(2:)=in%unit(axes+1:)
  end if
  call init_sep3d(out,"OUTPUT","FLOAT","REGULAR")
  call axis_allocate(out,size(n))
  out%n=n
  out%o=o
  out%d=d
  out%unit=unit
  out%label=label
  call sep3d_set_sep3d(out)
  call sep3d_write_description("out",out)
  if (verbose) then
    write(0,*) "Output dimensions ",out%n
  end if
  deallocate(n,o,d,label,unit)
  call sep_3d_close()
!################
!ALLOCATE MEMORY#
!################
  ntr=(maxsize*1000000)/4/out%n(1)
  nhelix=product(in%n(2:))
  ncompress_s=product(in%n(2:axes))
  nbig=nhelix/ncompress       !total number of output samples
  nbig_s=nbig
  ncompress_s=ncompress
  if (in%file_format.eq."REGULAR") then
    ntr=nbig_s*ncompress_s
  end if
  if (nbig.ne. ncompress .or. nbig.ne.nbig_s) then
    call seperr("the output data volume/n1 and compressing axes must&
      & both be less than 2gb")
  end if
  nblock_big= ntr / ncompress 
  !number of output samples we can do at one time
  if (nblock_big.eq.0) then
    call seperr("can't even process a single trace at a time")
  end if
  allocate(norm(nblock_big))
  if (verbose) then
    write(0,*) "processing ",nblock_big," output bins at a time"
  end if
  if (esize.eq.4) then
    allocate(input(in%n(1),ntr),output(out%n(1),nblock_big),stat=ierr)
  else
    allocate(inputc(in%n(1),ntr),outputc(out%n(1),nblock_big),stat=ierr&
      &)
  end if 
  if (ierr.ne.0) then
    call erexit("trouble allocating memory, try reducing maxsize")
  end if
  if (verbose) then
    write(0,*) "I will hold ",ntr," traces in memory"
  end if
!#################
! HELIXIZE INPUT #
!#################
  call axis_allocate(in,3)
  in%n(1:3)=(/out%n(1),ncompress_s,nbig_s/)
  call sep3d_set_sep3d(in)
!#################
!   LOOP DATA    #
!#################
  ipct_print=1
  tot=0
  ndone_big=0
  do while  ( ndone_big < nbig)
!loop over output space
    ndo_big=min(nbig-ndone_big, nblock_big)
    read=0
    norm=0
    if (esize.eq.4) then
      output=0
    else
      outputc=0.
    end if 
    do while (read < ncompress)
!LOOP UNTIL WE HAVE FOUND ALL THE TRACES TO COMPRESS
      block=min(ntr,ncompress-read)
!READ IN THE DATA
      if (in%file_format.ne."REGULAR") then
        if (nh>0) then
          call sep3d_grab_headers("in",in,nh,fwind=(/read,ndone_big/)&
            &,nwind=(/block,ndo_big/))
          allocate(coords(2,nh))
          if (.not. sep3d_grab_coords(in,coords)) then
            call seperr("trouble grabbing coordinates for input")
          end if
          if (esize.eq.4) then
            if (.not. sep3d_read_data("in",in,input(:,:nh))) then
              call erexit("trouble reading data")
            end if
          else if (esize.eq.8) then
            if (.not. sep3d_read_data("in",in,inputc(:,:nh))) then
              call erexit("trouble reading data")
            end if
          end if
          do i3=1,nh 
            norm(coords(2,i3)-ndone_big)=              norm(coords(2,i3&
              &)-ndone_big)+1
            if (esize.eq.4) then
              output(:,coords(2,i3)-ndone_big)= output(:,coords(2,i3)&
                &-ndone_big)+input(:,i3) !SUM THE TRACES
            else
              output(:,coords(2,i3)-ndone_big)=  output(:,coords(2,i3)&
                &-ndone_big)+inputc(:,i3)
            end if
          end do 
          deallocate(coords)
        end if
      else
        if (esize.eq.4) then
          if (esize*block*in%n(1).ne.sreed("in",input,esize*block*in%n&
            &(1))) then
            call erexit("trouble reading data")
          end if
        else
          if (esize*block*in%n(1).ne.sreed("in",inputc,esize*block*in%n&
            &(1))) then
            call erexit("trouble reading data")
          end if
        end if 
        nh=block
        norm=ncompress
!LOOK AT THIS LATER
!          if(esize==4) output+=sum(input(:,1:nh),2) #SUM THE TRACES
!          else outputc+=sum(inputc(:,1:nh),2)
      end if 
      read=      read+block !update number of traces read
    end do 
    if (verbose) then
      tot=        tot+sum( norm)
      ipct=100.*real(i3)/real(nbig)
      if (ipct >ipct_print) then
        ipct_print=           ipct_print+1
        write(0,*) "finished grid pct=",ipct,"trace print=",real(tot)&
          &/real(in%ntraces)*100.
      end if
    end if
    if (any(norm>0) .and. normalize) then
      do i3=1,ndo_big 
        if (norm(i3)>0) then
          if (esize.eq.4) then
            output(:,i3)=output(:,i3)/norm(i3)
          else
            outputc(:,i3)=outputc(:,i3) /norm(i3)
          end if
        end if
      end do
    end if
    if (esize.eq.4) then
      if (size(output)*esize.ne. srite("out",output,size(output)*esize&
        &)) then
        call        seperr("trouble writing out")
      end if
    else
      if (size(outputc)*esize.ne. srite("out",outputc,size(outputc)*esize&
        &)) then
        call        seperr("trouble writing out")
      end if
    end if 
    ndone_big=   ndone_big+ndo_big
  end do 
end program  
